import numpy as np
from colorama import Fore, Back, Style
default_tol=1e-12

success=f"{Back.GREEN}{Fore.BLACK}SUCCESS :){Style.RESET_ALL}"
failure=f"{Back.RED}{Fore.BLACK}ERROR   :({Style.RESET_ALL}"

def check_same(expected:np.ndarray,calculated:np.ndarray,tol=default_tol,check_shape=True):
    if expected is None:
        print(f"{failure} Expected value is None")
        return
    if calculated is None:
        print("{failure} Calculated value is None")
        return

    if check_shape:
        if not expected.shape ==calculated.shape:
            print(f"{failure} Shape mismatch:")
            print(f"Expected:   {expected.shape}")
            print(f"Calculated: {calculated.shape}")
            return
    if approximate_equals(expected,calculated,tol=tol):
        print(f"{success} Arrays are equal (tolerance {tol})")
    else:
        print(f"{failure} Arrays are not equal (tolerance {tol})")
        print(f"Expected:\n{expected}")
        print(f"Calculated:\n{calculated}")

def approximate_equals(a:np.ndarray,b:np.ndarray,tol=default_tol):
    return np.all(np.abs(a-b)<tol)


def check_different(a1:np.ndarray, a2:np.ndarray, tol=default_tol,check_shape=True):
    if check_shape:
        if not a1.shape ==a2.shape:
            print(f"{failure} Shape mismatch:")
            print(f"Expected:   {a1.shape}")
            print(f"Calculated: {a2.shape}")
            return

    if not approximate_equals(a1, a2, tol=tol):
        print(f"{success} Arrays are different  (tolerance {tol})")
    else:
        print(f"{failure} Arrays are not different  (tolerance {tol})")
        print(f"First:\n{a1}")
        print(f"Second:\n{a2}")


def check_same_shape(a1:np.ndarray, a2:np.ndarray):
    if a1.shape==a2.shape:
        print(f"{success} Arrays are have the same shape: {a1.shape}")
    else:
        print(f"{failure} Arrays have different shapes.")
        print(f"First:\n{a1.shape}")
        print(f"Second:\n{a2.shape}")

def check_same_float(expected:float,calculated:float,title:str,tol=default_tol):
    if np.abs((expected-calculated))<tol:
        print(f"{success} {title} is {expected}  (tolerance {tol})")
    else:
        print(f"{failure} {title} is not {expected}  (tolerance {tol})")
        print(f"Expected:   {expected} ")
        print(f"Calculated: {calculated}")

def check_mean(a:np.ndarray, mean:float, tol=default_tol):
    check_same_float(a.mean(),mean,"Mean",tol=tol)

def check_std(a:np.ndarray, std:float, tol=default_tol):
    check_same_float(a.std(),std,"Std",tol=tol)


from edunn.model import ModelWithParameters,Cache,ModelWithoutParameters
import numpy as np

class Identity(ModelWithParameters):

    def forward(self, x:np.ndarray):
        return x

    def backward(self,δEδy:np.ndarray):
        δEδp={} # no parameters, no derivatives
        return δEδy,δEδp

class AddConstant(ModelWithParameters):
    '''
    A layer that adds a constant
    This layer has NO parameters
    '''

    def __init__(self,value:float,name=None):
        super().__init__(name=name)
        self.value=value

    def forward(self, x:np.ndarray):
        '''
        :param x: input vector/matrix
        :return: x + a constant value, stored in self.value
        '''
        y= np.zeros_like(x)
        ### YOUR IMPLEMENTATION START  ###

        y=x+self.value
        ### YOUR IMPLEMENTATION END  ###
        return y

    def backward(self,δEδy:np.ndarray):
        δEδx= np.zeros_like(δEδy)
        ### YOUR IMPLEMENTATION START  ###
        δEδx= δEδy
        ### YOUR IMPLEMENTATION END  ###
        δEδp={} # no parameters, no derivatives
        return δEδx,δEδp

class MultiplyConstant(ModelWithParameters):
    '''
    A layer that multiplies by a constant
    This layer has NO parameters
    '''
    def __init__(self,value:float,name=None):
        super().__init__(name=name)
        self.value=value
    def forward(self, x:np.ndarray):
        '''
        :param x: input vector/matrix
        :return: x * a constant value, stored in self.value
        '''
        y= np.zeros_like(x)
        ### YOUR IMPLEMENTATION START  ###
        y=x*self.value
        ### YOUR IMPLEMENTATION END  ###
        return y


    def backward(self,δEδy:np.ndarray):
        δEδx= np.zeros_like(δEδy)

        ### YOUR IMPLEMENTATION START  ###
        δEδx=δEδy *self.value
        ### YOUR IMPLEMENTATION END  ###

        δEδp={} # no parameters, no derivatives
        return δEδx,δEδp



class ReLU(ModelWithoutParameters):

    def forward(self, x:np.ndarray):
        y = np.zeros_like(x)

        # TIP: NO utilizar np.max()
        # Ya que devuelve el valor máximo
        # y no aplica la función elemento a elemento

        ### YOUR IMPLEMENTATION START  ###
        y = np.maximum(x,0)
        ### YOUR IMPLEMENTATION END  ###
        self.set_cache(y)
        return y

    def backward(self, δEδy:np.ndarray):
        δEδx = np.zeros_like(δEδy)
        y, = self.get_cache()

        # TIP: δEδx = δEδy * δyδx
        # δyδx is 1 if the output was greater than 0
        # and 0 otherwise

        ### YOUR IMPLEMENTATION START  ###
        δyδx = y>0
        δEδx = δEδy * δyδx
        ### YOUR IMPLEMENTATION END  ###

        return δEδx,{}


class Sigmoid(ModelWithoutParameters):

    def forward(self, x:np.ndarray):
        y = np.zeros_like(x)
        ### YOUR IMPLEMENTATION START  ###
        y =   1.0/(1.0 + np.exp(-x))
        ### YOUR IMPLEMENTATION END  ###
        cache = (y,)
        self.set_cache(y)
        return y

    def backward(self, δEδy:np.ndarray):
        δEδx= np.zeros_like(δEδy)
        y, = self.get_cache()
        # TIP: δEδx = δEδy * δyδx
        # First calculate δyδx
        # then multiply by δEδy (provided)

        ### YOUR IMPLEMENTATION START  ###
        δyδx = y * (1.0-y)
        δEδx = δEδy * δyδx
        ### YOUR IMPLEMENTATION END  ###

        return δEδx,{}



class TanH(ModelWithoutParameters):
    def __init__(self,name=None):
        super().__init__(name=name)
        self.sigmoid=Sigmoid()

    def forward(self, x:np.ndarray):
        y= np.zeros_like(x)
        # TIP: TanH2 is simply sigmoid*2-1
        # we already defined self.sigmoid as a sigmod Layer
        # use it to simplify your implementation
        ### YOUR IMPLEMENTATION START  ###
        s = self.sigmoid.forward(2*x)
        y= s * 2 - 1
        ### YOUR IMPLEMENTATION END  ###
        return y # this layer's cache is the same as the sigmoid's cache

    def backward(self,δEδy:np.ndarray):
        δEδx= np.zeros_like(δEδy)
        # TIP: If TanH2 is simply sigmoid*2-1
        # Calculate derivative of TanH
        # in terms of derivative of sigmoid
        ### YOUR IMPLEMENTATION START  ###
        δEδx,δEδp =self.sigmoid.backward(δEδy)
        δEδx = δEδx*4
        ### YOUR IMPLEMENTATION END  ###

        return δEδx,{}


class Softmax(ModelWithoutParameters):
    def __init__(self,name=None,smoothing=1e-16):
        super().__init__(name)
        self.smoothing=smoothing

    def forward(self, x:np.ndarray):
        # add a small value so that no probability ends up exactly 0
        # This avoids NaNs when computing log(p) or 1/p
        # Specially when paired with the CrossEntropy error function
        x=x+self.smoothing

        n,classes=x.shape
        y = np.zeros_like(x)
        for i in range(n):
            xi=x[i,:]
            xi = xi + xi.max() # trick to avoid numerical issues
            # Calcular las probabilidades para cada clase
            # y guardar el valor en y[i,:]
            # en base al vector de puntaje xi
            # Nota: este cálculo es para 1 ejemplo del batch
            # el for se encarga de repetirlo para c/u
            ### YOUR IMPLEMENTATION START  ###
            e = np.exp(xi)
            N = e.sum()
            y[i,:] = e/N
            ### YOUR IMPLEMENTATION END  ###
        self.set_cache(y)
        return y

    def backward(self, δEδy:np.ndarray):
        # δEδx = δEδy * δyδx
        y, = self.get_cache()
        n,classes = δEδy.shape
        δEδx = np.zeros_like(δEδy)
        for i in range(n):
            δEδx[i,:]= self.backward_sample(δEδy[i,:],y[i,:])
        return δEδx,{}


    def backward_sample(self,δEδy:np.ndarray,y:np.ndarray):
        # AYUDA PARA EL CÁLCULO
        # http://facundoq.github.io/guides/softmax_derivada.html
        '''
        :param δEδy: derivative of error wrt output for a *single sample*
        :param y: output for a *single sample*
        :return: δEδx for a *single sample*
        '''
        classes=y.shape[0]

        # AYUDA PARA EL CÁLCULO
        # http://facundoq.github.io/guides/softmax_derivada.html
        δyδx = np.zeros((classes,classes))
        ### YOUR IMPLEMENTATION START  ###
        for i in range(classes):
            for j in range(classes):
                if i == j:
                    δyδx[i,j]=(1-y[i])*y[j]
                else:
                    δyδx[i,j]=-y[i]*y[j]

        # # Vectorized Version
        # id = np.identity(classes)
        # y = y[:,np.newaxis]
        # A = y.repeat(classes,axis=1)
        # B = y.T.repeat(classes,axis=0)
        # δyδx = (id-A)*B

        ### YOUR IMPLEMENTATION END  ###

        δEδx = np.zeros_like(δEδy)
        classes=y.shape[0]
        for j in range(classes):
            δyδx_j = δyδx[:,j]

            ### YOUR IMPLEMENTATION START  ###
            δEδx[j] = δEδy.dot(δyδx_j)
            ### YOUR IMPLEMENTATION END  ###
        return δEδx




from edunn.model import ModelWithParameters
import numpy as np
from edunn.initializers import Initializer, Zero


class Bias(ModelWithParameters):
    '''
    The Bias layer outputs y = x+b, where b is a vector of parameters
    Input for forward:
    x: array of size (n,o)
    where $n$ is the batch size and $o$ is the number of both input and output features
    The number of columns of x, $o$, must match the size of $b$.

    '''

    def __init__(self, output_size: int, initializer: Initializer = None, name=None):
        super().__init__(name=name)
        if initializer is None:
            initializer = Zero()
        b = initializer.create((output_size,))
        self.register_parameter("b", b)

    def forward(self, x: np.ndarray):
        n, d = x.shape
        b = self.get_parameters()["b"]
        dout, = b.shape
        assert dout == d, f"#features of input ({d}) must match size of b ({dout})"
        y = np.zeros_like(x)
        ### YOUR IMPLEMENTATION START  ###

        y = x + b

        ### YOUR IMPLEMENTATION END  ###
        return y

    def backward(self, δEδy: np.ndarray):
        b = self.get_parameters()["b"]
        δEδx = np.zeros_like(δEδy)

        # Calculate derivative of error E with respect to input x
        # Hints:
        # δEδx = δEδy * δyδx = δEδy * [1,1,...,1] = δEδy
        ### YOUR IMPLEMENTATION START  ###
        δEδx = δEδy
        ### YOUR IMPLEMENTATION END  ###

        # Calculate derivative of error E with respect to parameter b

        # Hints:
        # δEδb = δEδy * δyδb
        # δyδb = [1, 1, 1, ..., 1]
        n, d = δEδy.shape
        δEδb = np.zeros_like(b)
        for i in range(n):
            # Calculate derivative of error for a sample i (a single sample)
            # And accumulate to obtain δEδb
            ### YOUR IMPLEMENTATION START  ###
            δEδb_i = δEδy[i, :]  # * [1,1,1...,1]
            δEδb += δEδb_i
            ### YOUR IMPLEMENTATION END  ###

        return δEδx, {"b": δEδb}



from .sequential import Sequential
from .bias import Bias
from .linear import Linear
from .activations import  Sigmoid,TanH,ReLU,MultiplyConstant,AddConstant,Softmax

from .binary_cross_entropy_error import BinaryCrossEntropy
from .cross_entropy_error import CrossEntropyWithLabels
from .squared_error import SquaredError
from .dense import Dense
from .linear_regression import LinearRegression
from .logistic_regression import LogisticRegression
from .mean_error import MeanError




import numpy as np




def numerical_gradient(f,x:np.ndarray,δEδy:np.ndarray=None, h=1e-5):
    ''' Calculates the numerical gradient of E wrt x'''
    ''' E is assumed to be a scalar, so that δEδy has size equal to y'''
    def indices_generator(x):
        it = np.nditer(x, flags=['multi_index'], op_flags=['readwrite'])
        while not it.finished:
            # evaluate function at x+h
            ix = it.multi_index
            yield ix
            it.iternext()
    h2 = 2*h
    δEδx = np.zeros_like(x)
    for i in indices_generator(x):
        oldval = x[i]
        x[i] = oldval + h # increment by h
        fxph = f(x) # evaluate f(x + h)
        x[i] = oldval - h # increment by h
        fxmh = f(x) # evaluate f(x - h)
        x[i] = oldval # reset

        δyδxi = (fxph - fxmh)/h2

        δE = (δyδxi*δEδy)
        δE = δE.sum()
        δEδx[i] = δE
    return δEδx



import edunn as nn
import numpy as np
from typing import Tuple,Dict,Callable

ParameterSet = Dict[str,np.ndarray]

def set_parameters_inplace(l:nn.Model, parameters:ParameterSet):
    old_params = l.get_parameters()
    for k, v in old_params.items():
        l.get_parameters()[k][:]=parameters[k]

def common_layer_to_function(l:nn.ModelWithParameters):

    def f(inputs:Dict[str,np.ndarray]):
        old_params = l.get_parameters().copy()
        set_parameters_inplace(l,inputs)
        y = l.forward(inputs["x"])

        set_parameters_inplace(l,old_params)
        return y

    def df(inputs:ParameterSet,δEδy:np.ndarray):
        old_params = l.get_parameters().copy()
        set_parameters_inplace(l,inputs)
        y=l.forward(inputs["x"])
        δEδx,δEδp = l.backward(δEδy)
        set_parameters_inplace(l,old_params)
        return {"x":δEδx,**δEδp}
    parameter_shapes = {k:v.shape for k,v in l.get_parameters().items()}
    return f,df,parameter_shapes

def error_layer_to_function(l:nn.ErrorModel):

    def f(inputs:Dict[str,np.ndarray]):
        old_params = l.get_parameters().copy()
        set_parameters_inplace(l,inputs)
        y = l.forward(inputs["y_true"], inputs["y"])
        set_parameters_inplace(l,old_params)
        return y

    def df(inputs:Dict[str,np.ndarray]):
        old_params = l.get_parameters().copy()
        set_parameters_inplace(l,inputs)
        y = l.forward(inputs["y_true"], inputs["y"])
        δEδy,δEδp = l.backward(np.ones_like(y))
        set_parameters_inplace(l,old_params)
        return {"y":δEδy,**δEδp}
    parameter_shapes = {k:v.shape for k,v in l.get_parameters().items()}

    return f,df,parameter_shapes



import numpy as np
from . import basepath

classification_basepath=basepath/ "classification_data"

def load_classification_dataset(filename,classes):
    data=np.loadtxt(classification_basepath / filename,skiprows=1,delimiter=",")
    x = data[:,:-1]
    y = data[:,-1:]
    y=y.astype(int)
    y = np.squeeze(y)
    return x,y, classes


study2d_easy = lambda: load_classification_dataset("study2d_easy.csv", ["Failed", "Passed"])
study2d = lambda: load_classification_dataset("study2d.csv", ["Failed", "Passed"])
study1d = lambda: load_classification_dataset("study1d.csv", ["Failed", "Passed"])
iris = lambda: load_classification_dataset("iris.csv", ["setosa", "versicolor", "virginica"])

loaders = {
    "iris":iris,
    "study1d":study1d,
    "study2d":study2d,
    "study2d_easy":study2d_easy,
}


import numpy as np
from . import utils

def accuracy(y_true:np.ndarray,y_pred:np.ndarray)->float:
    check_label_array(y_true)
    check_label_array(y_pred)
    return np.mean(y_pred==y_true)

def rmse(y_true:np.ndarray,y_pred:np.ndarray)->float:
    return np.sqrt(((y_true-y_pred)**2).sum(axis=1)).mean()

def mae(y_true:np.ndarray,y_pred:np.ndarray)->float:
    return np.abs(y_true-y_pred).mean(axis=0).mean()


def check_label_array(y:np.ndarray):
    if  len(y.shape)!=1:
        raise ValueError(f"True labels array must have a single dimension (shape is {y.shape} instead)")
    if not np.issubdtype(y.dtype, np.integer):
        raise ValueError(f"True labels array must be of type int (type is {y.dtype} instead)")

def check_binary(y:np.ndarray):
    mi,ma=y.min(),y.max()
    if not (0<=mi and ma<=1):
        raise ValueError(f"Binary labels must have value 0 or 1, found values from {mi} to {ma} instead.")

def precision(y_true:np.ndarray,y_pred:np.ndarray)->float:
    check_label_array(y_true)
    check_label_array(y_pred)
    check_binary(y_true)
    check_binary(y_pred)
    n = len(y_true)
    pred_true_indices = y_pred == 1
    true_positives = np.logical_and(y_true == 1,y_pred ==1)

    return np.sum(true_positives)/np.sum(pred_true_indices)

def recall(y_true:np.ndarray,y_pred:np.ndarray)->float:
    check_label_array(y_true)
    check_label_array(y_pred)
    check_binary(y_true)
    check_binary(y_pred)


    true_indices = y_true == 1
    true_positives = np.logical_and(y_true == 1,y_pred ==1)

    return np.sum(true_positives)/np.sum(true_indices)

def fscore(y_true:np.ndarray,y_pred:np.ndarray)->float:
    check_label_array(y_true)
    check_label_array(y_pred)
    check_binary(y_true)
    check_binary(y_pred)
    p,r = precision(y_true,y_pred),recall(y_true,y_pred)
    return 2*p*r/(p+r)

def confusion(y_true:np.ndarray,y_pred:np.ndarray)->np.ndarray:
    check_label_array(y_true)
    check_label_array(y_pred)

    classes = y_true.max()+1
    c = np.zeros((classes,classes),dtype=np.int)
    n = len(y_true)
    for i in range(n):
        c[y_true[i],y_pred[i]]+=1
    return c

def classification_summary_onehot(y_true:np.ndarray,y_pred:np.ndarray):
    assert(len(y_true.shape)==2)
    assert(np.all(y_true.shape==y_pred.shape))
    y_true=utils.onehot2labels(y_true)
    y_pred=utils.onehot2labels(y_pred)
    classification_summary(y_true,y_pred)

def classification_summary(y_true:np.ndarray,y_pred:np.ndarray):
    classes = y_true.max()+1

    print(f"Accuracy: {accuracy(y_true,y_pred)} ({classes} classes)")
    if classes==2:
        p,r,f = precision(y_true,y_pred),recall(y_true,y_pred),fscore(y_true,y_pred)
        print(f"Precision: {p}")
        print(f"Recall: {r}")
        print(f"F-score: {f}")
    print(f"Confusion matrix: (rows true, columns pred)")
    c = confusion(y_true,y_pred)
    print(c)


def regression_summary(y_true:np.ndarray,y_pred:np.ndarray):
    print(f"RMSE {rmse(y_true,y_pred)}")
    print(f"MAE {mae(y_true,y_pred)}")


import numpy as np

from ..model import ErrorModel

class SquaredError(ErrorModel):
    def forward(self, y_true:np.ndarray, y:np.ndarray):
        delta = (y - y_true)

        n = y_true.shape[0]
        E = np.zeros((n,1))
        ### YOUR IMPLEMENTATION START  ###
        E = np.sum(delta** 2, axis=1,keepdims=True)
        ### YOUR IMPLEMENTATION END  ###
        self.set_cache(delta)
        return E

    def backward(self, δEδEi):
        delta, = self.get_cache()
        # Calculate error w.r.t
        # y (the output of the model) and not y_true (which is a fixed value)
        δEδy=np.zeros_like(delta)
        ### YOUR IMPLEMENTATION START  ###
        δEiδy = 2 * delta
        δEδy = δEiδy*δEδEi
        ### YOUR IMPLEMENTATION END  ###
        return δEδy,{}



import numpy as np
from ..model import ModelWithParameters

from ..initializers import Initializer

from .linear import Linear
from .bias import Bias
from .activations import Softmax


class LogisticRegression(ModelWithParameters):
    '''
    A LinearRegression model applies a linear and bias function, in that order, to an input, ie
    y = wx+b, where w and b are the parameters of the Linear and Bias models,

    '''

    def __init__(self, input_size: int, output_size: int,
                 linear_initializer: Initializer = None, bias_initializer: Initializer = None, name=None):
        self.output_size = output_size
        self.input_size = input_size

        self.linear = Linear(input_size, output_size, initializer=linear_initializer)
        self.bias = Bias(output_size, initializer=bias_initializer)
        self.softmax = Softmax()
        super().__init__(name=name)

    def forward(self, x: np.ndarray):
        # calculate and return softmax(bias(linear(x)))
        y = np.zeros((x.shape[0],self.output_size)) # default value

        ### YOUR IMPLEMENTATION START  ###
        y_linear = self.linear.forward(x)
        y_bias = self.bias.forward(y_linear)
        y = self.softmax.forward(y_bias)
        ### YOUR IMPLEMENTATION END  ###
        return y

    def backward(self, δEδy: np.ndarray):
        # Compute gradients for the parameters of the bias and linear models
        δEδbias, δEδlinear, δEδsoftmax = {}, {}, {}
        δEδx = np.zeros((δEδy.shape[0],self.input_size)) # default value

        ### YOUR IMPLEMENTATION START  ###
        δEδx_softmax, δEδsoftmax = self.softmax.backward(δEδy)
        δEδx_bias, δEδbias = self.bias.backward(δEδx_softmax)
        δEδx, δEδlinear = self.linear.backward(δEδx_bias)
        ### YOUR IMPLEMENTATION END  ###

        # combine gradients for parameters from linear and bias models
        # to obtain parameters for the Linear Regression (lr) model
        δEδlr = {**δEδbias, **δEδlinear, **δEδsoftmax}
        return δEδx, δEδlr

    def get_parameters(self):
        # returns the combination of parameters of linear and bias models
        # assumes these models don't use the same parameter names
        # ie: Linear has `w`, bias has `b`
        p_linear = self.linear.get_parameters()
        p_bias = self.bias.get_parameters()
        p = {**p_linear, **p_bias}
        return p



import numpy as np

from ..model import ErrorModel
import edunn as nn

class BinaryCrossEntropy(ErrorModel):

    '''
    Returns the CrossEntropy between two binary class distributions
    Receives a matrix of probabilities Nx1 for each sample, with the probability
     for class 1, ie if p=0.7, this indicates P(C=1) =0.7 and P(C=0)=0.3
    Receives a a vector of class labels, also one for each sample, with values either 0 or 1.

'''

    def forward(self, y_true:np.ndarray, y:np.ndarray):
        y_true = np.squeeze(y_true)
        assert(len(y_true.shape) == 1)
        assert y.min() >= 0

        n,c=y.shape

        E = np.zeros((n,1))
        ### YOUR IMPLEMENTATION START  ###
        for i in range(n):
            miss = y_true[i] * y[i] + (1 - y_true[i]) * (1 - y[i])
            if miss==0:
                miss += nn.eps
            E[i] = - np.log(miss)
        # print(error)
        ### YOUR IMPLEMENTATION END  ###
        assert np.all(np.squeeze(E).shape == y_true.shape)
        self.set_cache(y_true,y)
        return E

    def backward(self, δEδyi):
        y_true,y = self.get_cache()
        δEδy = np.zeros_like(y)
        n,classes = y.shape
        ### YOUR IMPLEMENTATION START  ###
        for i in range(n):
            miss = y_true[i] * y[i] - (1 - y_true[i]) * (1 - y[i])
            if miss==0:
                miss+=nn.eps
            δEδy[i] = - 1/miss
        ### YOUR IMPLEMENTATION END  ###

        return δEδy*δEδyi,{}




import edunn as nn
import numpy as np


class FakeModel(nn.ModelWithParameters):
    # Fake model with a single parameter array with a parameter items
    # Parameters always initialized as `parameter`
    # Derivative of parameters is always `gradient`
    # Derivative of input is always 0
    def __init__(self,parameter:np.ndarray,gradient:np.ndarray):
        super().__init__()
        self.parameter = parameter
        self.gradient = gradient
        assert (np.all(parameter.shape==gradient.shape))
        self.register_parameter("parameter", self.parameter)

    def forward(self, x:np.ndarray):
        return x

    def backward(self, δEδy:np.ndarray):
        δEδx = 0
        δEδp = {"parameter": self.gradient}
        return δEδx, δEδp


class FakeError(nn.ErrorModel):
    def __init__(self,error=1,derivative_value=1,name=None):
        super().__init__()
        self.error=error
        self.derivative_value=1

    # Fake error function without parameters
    # forward always returns self.error
    # backward always returns array with same shape as input
    # and all values set to self.derivative_values
    def forward(self, y, y_true):
        self.set_cache(y.shape)
        return self.error

    def backward(self, δE:float):
        shape, = self.get_cache()
        return np.ones(shape)*self.derivative_value, {}


import matplotlib.pyplot as plt
plt.style.use('bmh')

import numpy as np
import edunn as nn


def plot_history(history,error_name="Error",filepath=None):
    plt.figure()
    plt.plot(history)
    plt.xlabel("Epochs")
    plt.ylabel(error_name)
    plt.title("Training history")
    if filepath is None:
        plt.show()
    else:
        plt.savefig(filepath)
        plt.close()


def plot_model_dataset_1d(x:np.ndarray, y:np.ndarray, model:nn.Model, filepath=None):
    assert x.shape[0]==y.shape[0],"Input and output arrays must have the same number of samples"
    plt.figure()
    plt.scatter(x,y)
    y_pred=model.forward(x)
    plt.plot(x,y_pred)
    plt.xlabel("x")
    plt.ylabel("y")
    if filepath is None:
        plt.show()
    else:
        plt.savefig(filepath)
        plt.close()


def plot_model_dataset_2d_classification(x:np.ndarray, y:np.ndarray, model:nn.Model, filepath=None, detail=0.1, title=""):
    assert x.shape[1]==2,f"x must have only two input variables (received x with {x.shape[1]} dimensions)"

    plt.figure()
    # Plot decision regions
    x_min, x_max = x[:, 0].min() - 1, x[:, 0].max() + 1
    y_min, y_max = x[:, 1].min() - 1, x[:, 1].max() + 1
    xx, yy = np.meshgrid(np.arange(x_min, x_max, detail),
                         np.arange(y_min, y_max, detail))

    Z = np.c_[xx.ravel(), yy.ravel()]

    Z = model.forward(Z)
    Z = Z.argmax(axis=1)
    title = f"{title}: Regiones de cada clase"
    Z = Z.reshape(xx.shape)
    plt.contourf(xx, yy, Z, alpha=0.3)  # ,  cmap='RdBu')
    plt.colorbar()
    plt.title(title)
    plt.xlabel("x")
    plt.ylabel("y")
    # Plot data
    plt.scatter(x[:,0],x[:,1],c=y)
    if filepath is None:
        plt.show()
    else:
        plt.savefig(filepath)
        plt.close()




def plot_model_dataset_1d_classification(x,y,model,filepath=None):
    plt.figure()
    xaxis=np.array(range(x.shape[0]))
    plt.scatter(xaxis,x,c=y,label="samples")
    xx = np.linspace(x.min(),x.max(),100)
    yy = model.forward(xx)
    plt.plot(xx,yy,label="decision boundary")
    plt.xlabel("index")
    plt.ylabel("x")
    plt.legend()
    if filepath is None:
        plt.show()
    else:
        plt.savefig(filepath)
        plt.close()


def regression1d_predictions(y_true:np.ndarray,y_pred:np.ndarray,filepath=None):
    assert y_true.shape[0] == y_pred.shape[0], "Input and output arrays must have the same number of samples"
    plt.figure()
    plt.plot(range(len(y_true)), y_true, label="true value")
    plt.plot(range(len(y_true)), y_pred, label="prediction")
    plt.xlabel("Sample ID")
    plt.ylabel("Prediction")
    plt.legend()
    if filepath is None:
        plt.show()
    else:
        plt.savefig(filepath)
        plt.close()

def plot_activation_function(m:nn.Model, range=(-10, 10), backward=False):
    plot_activation_functions([m],range=range,backward=backward)

def plot_activation_functions(models:[nn.Model], range=(-10, 10), backward=False):
    start,end=range
    x = np.linspace(start,end,100)
    names = [m.__class__.__name__ for m in models]
    fig=plt.figure()
    ax = fig.gca()
    # spines_origin(ax)
    for m in models:
        y = m.forward(x)
        name = m.__class__.__name__
        if backward:
            E = np.ones_like(x)
            y,_ = m.backward(E)
            name = f"{name} derivative"

        plt.plot(x,y,label=name)
    plt.legend()
    names = ", ".join(names)
    if backward:
        title = f"Derivatives of {names}"
    else:
        title = f"Values of {names}"
    plt.title(title)

def spines_origin(ax):
    ax.grid(True)
    ax.spines['left'].set_position('zero')
    ax.spines['right'].set_color('none')
    ax.spines['bottom'].set_position('zero')
    ax.spines['top'].set_color('none')


import numpy as np

from edunn import ErrorModel, Model
from edunn.model import ParameterSet


class MeanError(ErrorModel):
    '''
    Converts a Model that converts
    an error function for each sample
    into a mean error function with a single scalar output
    which can represent the final error of a network
    '''

    def __init__(self, sample_error: Model, name=None):
        if name is None:
            name = f"Mean{sample_error.name}"
        super().__init__(name=name)
        self.sample_error_layer = sample_error

    def forward(self, y_true: np.ndarray, y: np.ndarray):
        E = 0
        ### YOUR IMPLEMENTATION START  ###
        Ei = self.sample_error_layer.forward(y_true, y)
        E = np.mean(Ei)
        ### YOUR IMPLEMENTATION END  ###
        n = y_true.shape[0]
        self.set_cache(n)
        return E

    def backward(self, δE: float = 1) -> (np.ndarray, ParameterSet):
        '''

        :param δE:Scalar to scale the gradients
                Needed to comply with Model's interface
                Default is 1, which means no scaling
        :param cache: calculated from forward pass
        :return:
        '''
        n, = self.get_cache()
        # Since we just calculate the mean over n values
        # and the mean is equivalent to multiplying by 1/n
        # the gradient is simply 1/n for each value
        δEδy = np.ones((n, 1)) / n
        # Return error gradient scaled by δE
        δEδy *= δE
        # Calculate gradient for each sample
        δEδy_sample, δEδp_sample = self.sample_error_layer.backward(δEδy)
        assert_message = f"sample_error_layer's gradient must have n values (found {δEδy_sample.shape[0]}, expected {n})"
        assert δEδy_sample.shape[0] == n, assert_message
        return δEδy_sample, δEδp_sample


from abc import ABC, abstractmethod
import numpy as np
from enum import Enum

class_counter = {}
from typing import Tuple, Dict


class Phase(Enum):
    Training = "Training"
    Test = "Test"


ParameterSet = Dict[str, np.ndarray]
Cache = Tuple


class Model(ABC):
    '''

    A Model can perform forward and backward operations
    Each model has a unique name. Also, it has a set of parameters that can be updated.
    The forward operation takes an input (x) and generates an output (y)
    The backward takes an error gradient for the output (δEδy) and generates error gradients for the input (δEδx) and parameters (δEδp)
    The backward operation must be defined explicitly and is not automatically derived from the forward.
    The model can be frozen, so that its parameters are not updated/optimized.
    '''

    def __init__(self, name=None):
        if name is None:
            class_name = self.__class__.__name__
            count = class_counter.get(class_name, 0)
            name = f"{class_name}_{count}"
            class_counter[class_name] = count + 1
        self.phase = Phase.Training

        self.name = name
        self.frozen = False
        self.cache = None

    def set_cache(self, *values):
        self.cache = values

    def get_cache(self):
        return self.cache

    def set_phase(self, phase: Phase):
        '''
        Indicate to the model the current Phase (training/tests/other) the model is in.
        This can be used to code different behaviors in training and testing phases.
        :param phase: phase to set
        '''
        self.phase = phase

    def n_parameters(self) -> int:
        '''
        :return: number of individual parameters of the network
        '''
        result = 0
        for p in self.get_parameters().values():
            result += p.size
        return result

    @abstractmethod
    def get_parameters(self):
        pass

    @abstractmethod
    def forward(self, *x) -> (np.ndarray):
        pass

    @abstractmethod
    def backward(self, δEδy: np.ndarray) -> (np.ndarray, ParameterSet):
        pass

    def __repr__(self):
        return f"{self.name}"


class ModelWithParameters(Model):
    '''
    Helper class to implement layers with parameters
    '''

    def __init__(self, name=None):
        super().__init__(name=name)
        self.parameters = {}

    def register_parameter(self, name, value):
        self.parameters[name] = value

    def get_parameters(self):
        return self.parameters


class ModelWithoutParameters(Model):
    '''
    Helper class to implement layers _without_ parameters
    '''

    def __init__(self, name=None):
        super().__init__(name=name)

    def get_parameters(self):
        return {}


class ErrorModel(ModelWithoutParameters):
    '''
    Helper class to implement layers _without_ parameters
    '''

    def __init__(self, name=None):
        super().__init__(name=name)

    @abstractmethod
    def forward(self, y_true: np.ndarray, y: np.ndarray):
        pass

    @abstractmethod
    def backward(self, δE: float) -> (np.ndarray, ParameterSet):
        pass





import numpy as np
from edunn.model import Model, Phase


class Sequential(Model):
    '''
        Models a neural network with a sequential (ie, linear) topology
        This network receives as input a single vector x
        And outputs a single vector y
    '''

    def __init__(self, layers: [Model], name=None):
        '''

        :param layers: List of models, in order

        '''
        super().__init__(name)
        self.layers: [Model] = layers
        layer_names = [l.name for l in layers]
        layer_names_set = set(layer_names)
        assert len(layer_names) == len(layer_names_set), f"Layer names must be unique, found: {layer_names}"

    def forward(self, x: np.ndarray):
        '''

        :param x: input to model
        :return: output of model with x as input
        '''
        y = 0
        ### YOUR IMPLEMENTATION START  ###
        for layer in self.layers:
            x = layer.forward(x)
        y = x
        ### YOUR IMPLEMENTATION END  ###
        return y

    def set_phase(self, phase: Phase):
        for l in self.layers:
            l.set_phase(phase)

    def merge_gradients(self, m_i, δEδp_i, δEδp):
        """
        :param m_i: model for ith layer
        :param δEδp_i: derivatives parameters of model for ith layer
        :param δEδp: all derivatives of parameters of Sequential
        """

        if not m_i.frozen:
            for k, v in δEδp_i.items():
                new_name = self.generate_parameter_name(m_i, k)
                δEδp[new_name] = v

    def backward(self, δEδy: np.ndarray):
        '''

        :param x: inputs
        :param y: expected output
        :return: gradients for every layer, prediction for inputs and error
        '''
        δEδx = 0
        δEδp = {}
        # Hint: use `self.merge_gradients(m_i, δEδp_i, δEδp)`
        # to add the gradients `δEδp_i` of parameters of `m_i`
        # to the final gradients `δEδp` dictionary
        ### YOUR IMPLEMENTATION START  ###
        for m_i in reversed(self.layers):
            δEδy, δEδp_i = m_i.backward(δEδy)
            self.merge_gradients(m_i, δEδp_i, δEδp)
        δEδx = δEδy
        ### YOUR IMPLEMENTATION END  ###

        return δEδx, δEδp

    def generate_parameter_name(self, l: Model, parameter_name: str):
        return f"{l.name}({parameter_name})"

    def get_parameters(self):
        parameters = {}
        for l in self.layers:
            for k, v in l.get_parameters().items():
                new_name = self.generate_parameter_name(l, k)
                parameters[new_name] = v
        return parameters

    def summary(self) -> str:
        '''
        :return: a summary of the models of the model and their parameters
        '''
        separator = "-------------------------------"
        result = f"{separator}\n"
        parameters = 0
        result += f"Model {self.name}:\n"
        for l in self.layers:
            layer_parameters = l.n_parameters()
            parameters += layer_parameters
            l_summary = f"{l.name} → params: {layer_parameters}"
            result += l_summary + "\n"
        result += f"Total parameters: {parameters}\n"
        result += f"{separator}\n"
        return result



from .comparisons import *
from . import check_gradient
from .numerical_gradient import numerical_gradient

import numpy as np

def onehot2labels(onehot):
    return onehot.argmax(axis=1)

def labels2onehot(labels,n_classes):
    n = len(labels)
    onehot = np.zeros((n, n_classes))
    for i in range(n):
        onehot[i,labels[i]] = 1
    return onehot


import numpy as np
import pathlib
from typing import Dict,Callable
basepath = pathlib.Path(__file__).parent.absolute()
from .. import utils
from . import regression,classification

def shuffle_dataset(x,y):
    n = x.shape[0]
    indices = np.random.permutation(n)
    y = y[indices,:]
    x = x[indices,:]
    return x,y


def load_regression(dataset_name:str):
    return load(dataset_name,regression.loaders,"regression")

def load_classification(dataset_name:str,onehot=False):
    x,y,classes= load(dataset_name,classification.loaders,"classification")
    if onehot:
        y = utils.labels2onehot(y,len(classes))
    return x,y,classes

def load(dataset_name:str,loaders:Dict[str,Callable],title:str):
    if not dataset_name in loaders:
        raise ValueError(f"Unknown dataset {dataset_name}. Valid choices for {title} datasets:\n {loaders.keys()}.")

    dataset_loader = loaders[dataset_name]
    return dataset_loader()


def get_classification_names():
    return classification.loaders.keys()

def get_regression_names():
    return regression.loaders.keys()


eps = 1e-12

from .model import ErrorModel,Model,Phase, ModelWithParameters
from .optimizer import Optimizer,BatchedGradientOptimizer,GradientDescent,MomentumGD,NesterovMomentumGD,SignGD
from . import initializers,plot

from .models import *
from . import metrics,datasets


from edunn.models.fake import FakeModel,FakeError



import numpy as np

from ..model import ErrorModel


class CrossEntropyWithLabels(ErrorModel):
    '''
        Returns the CrossEntropy between two class distributions
        Receives a matrix of probabilities NxC for each sample and a vector
        of class labels, also one for each sample,.
        For each sample, this layer receives a vector of probabilities for each class,
        (which must sum to 1)
        and the label of the sample (which implies that class has an
        expected probability of 1, and 0 for the rest)
    '''

    ### Ayuda para implementar:
    ### http://facundoq.github.io/guides/crossentropy_derivative.html
    def forward(self, y_true:np.ndarray, y:np.ndarray):
        y_true = np.squeeze(y_true)
        assert len(y_true.shape) == 1
        assert y.min() >= 0
        n,c=y.shape

        E = np.zeros((n,1))
        ### YOUR IMPLEMENTATION START  ###
        for i in range(n):
            probability = y[i, y_true[i]]
            E[i] = -np.log(probability)
        ### YOUR IMPLEMENTATION END  ###
        assert np.all(np.squeeze(E).shape == y_true.shape)
        self.set_cache(y_true,y)
        return E

    def backward(self, δEδyi):
        y_true,y = self.get_cache()

        δEδy = np.zeros_like(y)
        n,classes = y.shape
        ### YOUR IMPLEMENTATION START  ###
        for i in range(n):
            p=y[i, y_true[i]]
            δEδy[i, y_true[i]] = -1 / p
        ### YOUR IMPLEMENTATION END  ###
        return δEδy*δEδyi,{}



import numpy as np
from . import basepath

regression_basepath=basepath/ "regression_data"
def load_regression_dataset(filename,output_values=1):
    data=np.loadtxt(regression_basepath / filename,skiprows=1,delimiter=",")
    x = data[:,:-output_values]
    y = data[:,-output_values:]
    return x,y

#https://www.kaggle.com/c/house-prices-advanced-regression-techniques/data?select=train.csv
# TODO cleanup
houses = lambda: load_regression_dataset("houses.csv")

study1d = lambda: load_regression_dataset("study1d.csv")
study2d = lambda: load_regression_dataset("study2d.csv",output_values=2)

#https://www.kaggle.com/c/boston-housing
boston = lambda: load_regression_dataset("boston_housing.csv")

#https://archive.ics.uci.edu/ml/datasets/wine+quality
wine_red = lambda: load_regression_dataset("wine_red.csv")
wine_white = lambda: load_regression_dataset("wine_white.csv")

#https://www.kaggle.com/quantbruce/real-estate-price-prediction
real_state = lambda: load_regression_dataset("real_state.csv")

#https://www.kaggle.com/mirichoi0218/insurance
insurance = lambda: load_regression_dataset("insurance.csv")

loaders = {
    "study1d":study1d,
    "study2d":study2d,
    "boston":boston,
    "wine_red":wine_red,
    "wine_white":wine_white,
    "insurance":insurance,
    "real_state":real_state,
}


import abc
import numpy as np
from typing import Tuple


class Initializer(abc.ABC):

    @abc.abstractmethod
    def initialize(self,p:np.ndarray):
        pass

    def create(self,shape:Tuple[int,...]):
        p = np.zeros(shape)
        self.initialize(p)
        return p

class Zero(Initializer):
    def initialize(self,p:np.ndarray):
        p[:]=0


class Constant(Initializer):
    def __init__(self,c:np.ndarray):
        self.c=c

    def initialize(self,p:np.ndarray):
        ### YOUR IMPLEMENTATION START  ###
        p[:]=self.c
        ### YOUR IMPLEMENTATION END  ###


class RandomUniform(Initializer):
    '''
    A random initializer with samples from a Uniform distribution
    with mean 0 and range -a,a
    '''
    def __init__(self, a:float=1e-10):
        super().__init__()
        self.a=a

    def initialize(self,p:np.ndarray):
        #TIP use np.random.uniform(a,b,shape)
        #to generate uniform random numbers between a and b

        ### YOUR IMPLEMENTATION START  ###
        p[:]=np.random.uniform(-self.a,self.a,p.shape)
        ### YOUR IMPLEMENTATION END  ###


class RandomNormal(Initializer):
    def __init__(self, std:float=1e-10):
        super().__init__()
        self.std=std

    def initialize(self,p:np.ndarray):
        #TIP use np.random.normal(μ,σ,shape)
        # to generate random numbers sampled from
        # a normal distribution with mean μ and std σ

        ### YOUR IMPLEMENTATION START  ###
        p[:]=np.random.normal(0,self.std,p.shape)
        ### YOUR IMPLEMENTATION END  ###


class XavierUniform(Initializer):

    def initialize(self,p:np.ndarray):
        din,dout=p.shape
        #TIP use np.random.random_sample(shape)
        # to generate uniform randomnumbers between 0 and 1
        ### YOUR IMPLEMENTATION START  ###
        factor = np.sqrt(6/(din+dout))
        p[:]=(np.random.random_sample(p.shape)*2-1)*factor
        ### YOUR IMPLEMENTATION END  ###

class XavierNormal(Initializer):

    def initialize(self,p:np.ndarray):
        din,dout=p.shape
        #TIP use np.random.normal(mu,std,shape)
        # to sample random numbers from a normal distribution

        ### YOUR IMPLEMENTATION START  ###
        std = np.sqrt(6/(din+dout))
        p[:]=(np.random.normal(0,std, p.shape))
        ### YOUR IMPLEMENTATION END  ###

class KaimingNormal(Initializer):

    def initialize(self,p:np.ndarray):
        din,dout=p.shape
        #TIP use np.random.normal(mu,std,shape)
        # to sample random numbers from a normal distribution

        ### YOUR IMPLEMENTATION START  ###
        std = np.sqrt(2/din)
        p[:]=(np.random.normal(0,std, p.shape))
        ### YOUR IMPLEMENTATION END  ###

class KaimingUniform(Initializer):

    def initialize(self,p:np.ndarray):
        din,dout=p.shape
        #TIP use np.random.random_sample(shape)
        # to generate uniform randomnumbers between 0 and 1

        ### YOUR IMPLEMENTATION START  ###
        factor = np.sqrt(2/din)
        p[:]=(np.random.random_sample(p.shape)*2-1)*factor
        ### YOUR IMPLEMENTATION END  ###


import numpy as np
from typing import Tuple,Dict,Callable
import edunn as nn


from edunn.utils.model2function import common_layer_to_function,error_layer_to_function,ParameterSet
from edunn.utils.numerical_gradient import numerical_gradient
from colorama import Fore, Back, Style



def common_layer(l:nn.ModelWithParameters, x_shape:Tuple, samples:int=1, tolerance=1e-7, break_on_error=True):
    f,df,parameter_shapes = common_layer_to_function(l)
    shapes = {**parameter_shapes,"x":x_shape}
    input_generators = lambda: {k:np.random.normal(0,1,shape) for k,shape in shapes.items()}
    y = f(input_generators())
    δEδy_generator = lambda: np.random.normal(0,1,y.shape)
    common_layer_random_sample(l, f, df, input_generators, δEδy_generator, samples=samples, tolerance=tolerance, break_on_error=break_on_error)

def squared_error(l:nn.SquaredError, y_shape:Tuple, samples:int=1, tolerance=1e-7, break_on_error=True):
    f,df,parameter_shapes = error_layer_to_function(l)

    df_ignore_δEδy = lambda x,δEδy: df(x)
    std=10
    def input_generators():
        parameters = {k:np.random.normal(0,1,shape) for k,shape in parameter_shapes.items()}
        y = np.random.normal(0,std,y_shape)
        y_true = np.random.normal(0,std,y_shape)
        return {**parameters,"y":y,"y_true":y_true}
    E = f(input_generators())
    δE_generator = lambda : np.ones(E.shape)

    common_layer_random_sample(l, f, df_ignore_δEδy, input_generators, δE_generator, samples=samples, tolerance=tolerance, break_on_error=break_on_error)

def cross_entropy_labels(l:nn.CrossEntropyWithLabels, y_shape:Tuple, samples:int=1, tolerance=1e-7, break_on_error=True):
    f,df,parameter_shapes = error_layer_to_function(l)
    df_ignore_δEδy = lambda x,δEδy: df(x)

    def input_generators():
        parameters = {k:np.random.normal(0,1,shape) for k,shape in parameter_shapes.items()}

        y = np.abs(np.random.normal(0,1,y_shape))
        y/=y.sum(axis=1,keepdims=True)


        n,classes = y.shape
        y_true=np.random.randint(0,classes,(n,))

        return {**parameters,"y":y,"y_true":y_true}

    E = f(input_generators())
    δE_generator = lambda : np.ones(E.shape)
    common_layer_random_sample(l, f, df_ignore_δEδy, input_generators, δE_generator, samples=samples, tolerance=tolerance, break_on_error=break_on_error)

def binary_cross_entropy_labels(l:nn.BinaryCrossEntropy, batch_size:int, samples:int=1, tolerance=1e-7, break_on_error=True):
    f,df,parameter_shapes = error_layer_to_function(l)
    df_ignore_δEδy = lambda x,δEδy: df(x)
    sm=nn.Softmax()
    def input_generators():
        parameters = {k:np.random.normal(0,1,shape) for k,shape in parameter_shapes.items()}
        y = np.random.uniform(0,1,(batch_size,1))

        y_true=np.random.randint(0,1,(batch_size,1))

        return {**parameters,"y":y,"y_true":y_true}

    E = f(input_generators())
    δE_generator = lambda : np.ones(E.shape)
    common_layer_random_sample(l, f, df_ignore_δEδy, input_generators, δE_generator, samples=samples, tolerance=tolerance, break_on_error=break_on_error)


def common_layer_random_sample(l:nn.Model, f, df, input_generator, δEδy_generator, samples:int=1, tolerance=1e-7, break_on_error=True):
    checks,errors=0,0
    print(f"{Back.LIGHTBLUE_EX}{Fore.BLACK}{l.name} layer:{Style.RESET_ALL}")

    for i in range(samples):
        inputs = input_generator()
        δEδy = δEδy_generator()
        δEδinputs_analytic = df(inputs,δEδy=δEδy)
        sample_checks,sample_errors=check_gradient_numerical(f,inputs,δEδy,δEδinputs_analytic,tolerance=tolerance,break_on_error=break_on_error)
        errors+=sample_errors
        checks+=sample_checks
        if errors>0 and break_on_error:
            break

    if errors==0:
        status = f"{Back.GREEN}{Fore.BLACK}SUCCESS{Style.RESET_ALL}"
        print(f"{status} {checks} partial derivatives checked ({samples} random input samples)")

    if not break_on_error and errors>0:
        status = f"{Back.RED}{Fore.BLACK}ERROR{Style.RESET_ALL}"
        print(f"{status} {checks} partial derivatives checked, {errors} failed (tolerance {tolerance}, {samples} random input samples)")



def check_gradient_numerical(f:Callable,inputs:ParameterSet,δEδy:np.ndarray,δEδinputs_analytic:Dict[str,np.ndarray], tolerance:float,break_on_error:bool):
    errors=0
    checks=0
    for k,δEδk_analytical in δEδinputs_analytic.items():
        v  = inputs[k]
        #numerical
        def fk(x):
            old =inputs[k]
            inputs[k] = x
            y = f(inputs)
            inputs[k] = old
            return y
        δEδk_numerical = numerical_gradient(fk,v,δEδy)

        #comparison
        error_count,count,max_error=relative_error_count(δEδk_analytical,δEδk_numerical,tolerance=tolerance)
        errors+=error_count
        checks+= count
        if break_on_error and error_count>0 :

            report_errors(δEδk_analytical,δEδk_numerical,k,v,δEδy,max_error,tolerance,)

    return checks,errors


def report_errors(δEδk_analytical,δEδk_numerical,label:str,x:np.ndarray,δEδy:np.ndarray,max_error:float,tolerance:float):

    message = f"{Back.RED}{Fore.BLACK} ERROR {Style.RESET_ALL}"
    message+= f"\nδEδ{label}"
    message+= f"\n Relative error (max):{max_error:0.5f} (tolerance: {tolerance})"
    message+= f"\n{Style.RESET_ALL}######################## Details: ######################## "
    message+= f"\n Input {label}:\n{x}"
    message+= f"\n Input δEδy:\n{δEδy}"
    message+= f"\n δEδ{label} (numerical, automatic):\n{δEδk_numerical}"
    message+= f"\n δEδ{label} (analytic, your implementation):\n{δEδk_analytical}"

    message+= f"\n##########################################################\n{Style.RESET_ALL}"
    print(message)

def relative_error_count(δanalytical:np.ndarray, δnumerical:np.ndarray, tolerance=1e-8):
    """
    Check numerical gradient vs analytic gradient
    Count how many partial derivatives have a rel error greater than @tolerance
    """

    δsum= abs(δnumerical) + abs(δanalytical)
    δdiff=abs(δnumerical - δanalytical)
    rel_error = np.zeros_like(δsum)
    non_zero_indices = δsum!=0
    rel_error[non_zero_indices]=δdiff[non_zero_indices] / δsum[non_zero_indices]

    error_count = (rel_error>tolerance).sum()
    if error_count==0:
        max_error=0
    else:
        max_error=rel_error.max()

    return error_count,rel_error.size,max_error



# Additional material to help you implement optimizers:
# "An overview of gradient descent optimization algorithms" https://ruder.io/optimizing-gradient-descent/


from typing import Dict
import numpy as np
from .model import Model,Phase
from .model import ErrorModel, ParameterSet
import sys,abc
from tqdm.auto import tqdm

class Optimizer(abc.ABC):

    @abc.abstractmethod
    def optimize(self, model:Model, *args):
        pass




def all_equal(list:[]):
    return len(list) == 0 or list.count(list[0]) == len(list)
import random

def batch_arrays(batch_size:int,*arrays,shuffle=False):
    '''

    :param batch_size: size of batches
    :param arrays: variable number of numpy arrays
    :return: a generator that returns the arrays in batches
    '''

    sample_sizes = [a.shape[0] for a in arrays]
    assert all_equal(sample_sizes)
    batches=sample_sizes[0]//batch_size
    batch_list = list(range(batches))
    if shuffle:
        random.shuffle(batch_list)
    for i in batch_list:
        start = i*batch_size
        end = start+batch_size
        batch = [ a[start:end,] for a in arrays]
        yield tuple(batch)


class BatchedGradientOptimizer(Optimizer):

    def __init__(self,batch_size:int,epochs:int,shuffle=True):
        '''
        :param epochs: number of epochs to train the model. Each epoch is a complete iteration over the training set. The number of parameter updates is n //batch_size, where n is the number of samples of the dataset
        :param batch_size: Batch the dataset with batches of size `batch_size`, and perform an optimization step for each batch
        '''
        self.batch_size=batch_size
        self.epochs=epochs
        self.shuffle=shuffle

    def backpropagation(self,model:Model, x:np.ndarray, y_true:np.ndarray, error_layer:ErrorModel):
        # forward pass (model and error)
        y = model.forward(x)
        E = error_layer.forward(y_true, y)

        # backward pass (error and model)
        δEδy, _ = error_layer.backward(1)
        δEδx,δEδps = model.backward(δEδy)

        return δEδx,δEδps,E

    def optimize(self, model:Model, x:np.ndarray, y:np.ndarray, error_layer:ErrorModel, verbose=True):
        '''
        Fit a model to a dataset.
        :param model: the Model to optimize
        :param x: dataset inputs
        :param y: dataset outputs
        :param error_layer: To be applied to the output of the last layer
        :return:
        '''
        n = x.shape[0]
        batches = n // self.batch_size
        history = []
        model.set_phase(Phase.Training)
        bar = tqdm(range(self.epochs),desc=f"optim. {model.name}",file=sys.stdout,disable=not verbose)
        for epoch in bar:
            epoch_error=0
            for i,(x_batch,y_batch) in enumerate(batch_arrays(self.batch_size,x,y,shuffle=self.shuffle)):
                δEδx, δEδps, batch_error = self.backpropagation(model, x, y, error_layer)
                self.optimize_batch(model,δEδps,epoch,i)
                epoch_error+=batch_error
            epoch_error/=batches
            history.append(epoch_error)
            bar.set_postfix_str(f"{error_layer.name}: {epoch_error:.5f}")

        return np.array(history)

    @abc.abstractmethod
    def optimize_batch(self, model:Model, x:np.ndarray, y:np.ndarray, error_layer:ErrorModel, epoch:int):
        pass

class GradientDescent(BatchedGradientOptimizer):

    def __init__(self,batch_size:int,epochs:int,lr:float=0.1,shuffle=True):
        super().__init__(batch_size,epochs,shuffle)
        self.lr=lr

    def optimize_batch(self, model:Model, δEδps:ParameterSet, epoch:int,iteration:int):

        # Update parameters
        parameters = model.get_parameters()
        for parameter_name,δEδp in δEδps.items():
                p = parameters[parameter_name]
                # use p[:] so that updates are in-place
                # instead of creating a new variable
                ### YOUR IMPLEMENTATION START  ###
                p[:] = p - self.lr * δEδp
                ### YOUR IMPLEMENTATION END  ###

class MomentumGD(BatchedGradientOptimizer):

    def __init__(self,batch_size:int,epochs:int,lr:float=0.1,gamma=0.9,shuffle=True):
        super().__init__(batch_size,epochs,shuffle)
        self.lr=lr
        self.gamma=gamma
        self.first=True
        self.v={}

    def optimize_batch(self, model:Model, δEδps:ParameterSet, epoch:int,iteration:int):
        if self.first:
            self.first=False
            for k,p in model.get_parameters().items():
                self.v[k]=np.zeros_like(p)

        # Update parameters
        parameters = model.get_parameters()
        for k,δEδp in δEδps.items():
            # K = parameter name
            p = parameters[k]
            v = self.v[k]
            # use p[:] and v[:] so that updates are in-place
            # instead of creating a new variable
            ### YOUR IMPLEMENTATION START  ###
            v[:] = self.gamma * v + self.lr * δEδp
            p[:] = p - v
            ### YOUR IMPLEMENTATION END  ###


class NesterovMomentumGD(BatchedGradientOptimizer):

    def __init__(self,batch_size:int,epochs:int,lr:float=0.1,gamma=0.9,shuffle=True):
        super().__init__(batch_size,epochs,shuffle)
        self.lr=lr
        self.gamma=gamma
        self.first=True
        self.v={}

    def optimize_batch(self, model:Model, δEδps:ParameterSet, epoch:int,iteration:int):
        if self.first:
            self.first=False
            for k,p in model.get_parameters().items():
                self.v[k]=np.zeros_like(p)

        # Update parameters
        parameters = model.get_parameters()
        for k,δEδp in δEδps.items():
            # K = parameter name
            p = parameters[k]
            v = self.v[k]
            # use p[:] so that updates are in-place
            # instead of creating a new variable
            ### YOUR IMPLEMENTATION START  ###
            v[:] = self.gamma * v + self.lr * δEδp
            p[:] = p - (self.gamma*v+self.lr*δEδp)
            ### YOUR IMPLEMENTATION END  ###


class SignGD(BatchedGradientOptimizer):

    def __init__(self,batch_size:int,epochs:int,lr:float=0.1,eps=1e-8,shuffle=True):
        super().__init__(batch_size,epochs,shuffle)
        self.eps=eps
        self.lr=lr

    def optimize_batch(self, model:Model, δEδps:ParameterSet, epoch:int,iteration:int):

        # Update parameters
        parameters = model.get_parameters()
        for parameter_name,δEδp in δEδps.items():
                p = parameters[parameter_name]
                # use p[:] so that updates are in-place
                # instead of creating a new variable
                ### YOUR IMPLEMENTATION START  ###
                denom = np.sqrt(δEδp**2+self.eps)
                p[:] = p - self.lr * (δEδp/denom)
                ### YOUR IMPLEMENTATION END  ###




import numpy as np
from ..model import ModelWithParameters

from ..initializers import  Initializer,Zero,RandomNormal

from .linear import Linear
from . import activations
from .bias import Bias

activation_dict = {"id": activations.Identity,
                   "relu": activations.ReLU,
                   "tanh": activations.TanH,
                   "sigmoid": activations.Sigmoid,
                   "softmax": activations.Softmax,
                   }

class Dense(ModelWithParameters):
    '''
    A Dense layer simplifies the definition of networks by producing a common block
    that applies a linear, bias and activation function, in that order, to an input, ie
    y = activation(wx+b), where w and b are the parameters of the Linear and Bias models,
    and activation is the function of an activation Layer.

    Therefore, a defining a Dense layer such as:

    ```
    [...
    Dense(input_size,output_size,activation_name="relu")
    ]
    ```

    Is equivalent to:

    ```[...
    Linear(input_size,output_size),
    Bias(output_size)
    ReLu(),...]
    ```

    By default, no activation is used (actually, the Identity activation is used, which
    is equivalent). Implemented activations:
    * id
    * relu
    * tanh
    * sigmoid
    * softmax
    '''
    def __init__(self, input_size:int, output_size:int,activation_name:str=None,
                 linear_initializer:Initializer=None, bias_initializer:Initializer=None, name=None):
        self.linear = Linear(input_size,output_size,initializer=linear_initializer)
        self.bias = Bias(output_size,initializer=bias_initializer)

        if activation_name is None:
            activation_name = "id"
        if activation_name in activation_dict:
            self.activation = activation_dict[activation_name]()
        else:
            raise ValueError(f"Unknown activation function {activation_name}. Available activations: {','.join(activation_dict.keys())}")

        super().__init__(name=name)
        # add activation name to Dense name
        self.name+=f"({activation_name})"

    def forward(self, x:np.ndarray):
        # calculate and return activation(bias(linear(x)))
        y_activation = None
        ### YOUR IMPLEMENTATION START  ###
        y_linear = self.linear.forward(x)
        y_bias =self.bias.forward(y_linear)
        y_activation= self.activation.forward(y_bias)
        ### YOUR IMPLEMENTATION END  ###
        return y_activation

    def backward(self,δEδy:np.ndarray):
        # Compute gradients for the parameters of the bias, linear and activation function
        # It is possible that the activation function does not have any parameters
        # (ie, δEδactivation = {})
        δEδbias,δEδlinear,δEδactivation={},{},{}
        δEδx = None
        ### YOUR IMPLEMENTATION START  ###
        δEδx_activation,δEδactivation = self.activation.backward(δEδy)
        δEδx_bias,δEδbias =self.bias.backward(δEδx_activation)
        δEδx,δEδlinear =self.linear.backward(δEδx_bias)
        ### YOUR IMPLEMENTATION END  ###

        # combine gradients for parameters from dense, linear and activation models
        δEδdense ={**δEδbias, **δEδlinear,**δEδactivation}
        return δEδx,δEδdense


    def get_parameters(self):
        # returns the combination of parameters of all models
        # assumes no Layer uses the same parameter names
        # ie: Linear has `w`, bias has `b` and activation
        # has a different parameter name (if it has any).
        p_linear = self.linear.get_parameters()
        p_bias = self.bias.get_parameters()
        p_activation = self.activation.get_parameters()
        p = {**p_linear, **p_bias,**p_activation}
        return p





from edunn.model import ModelWithParameters
import numpy as np
from edunn.initializers import  Initializer,RandomNormal

class Linear(ModelWithParameters):

    '''
    The Linear layer outputs y = xw, where w is a matrix of parameters

    '''
    def __init__(self,input_size:int,output_size:int,initializer:Initializer=None,name=None):
        super().__init__(name=name)
        if initializer is None:
            initializer = RandomNormal()
        shape = (input_size,output_size)
        w = initializer.create(shape)
        self.register_parameter("w", w)


    def forward(self, x:np.ndarray):
        n,d = x.shape
        # Retrieve w
        w = self.get_parameters()["w"]
        # check sizes
        din,dout = w.shape
        assert din==d, f"#features of input ({d}) must match first dimension of W ({din})"



        y = np.zeros((n,dout))
        # calculate output
        ### YOUR IMPLEMENTATION START  ###
        y = x.dot(w)
        ### YOUR IMPLEMENTATION END  ###

        # add input to cache to calculate δEδw in backward step
        self.set_cache(x)
        return y

    def backward(self,δEδy:np.ndarray):
        # Retrieve input from cache to calculate δEδw
        x, = self.get_cache()
        n = x.shape[0]

        # Retrieve w
        w = self.get_parameters()["w"]

        # Calculate derivative of error E with respect to input x
        δEδx = np.zeros_like(x)
        ### YOUR IMPLEMENTATION START  ###

        # Per sample version
        # for i in range(n):
        #      δEδx[i,:] = np.dot(w, δEδy[i,:])

        # Vectorized version
        #δyδx = w.T
        δEδx =δEδy.dot(w.T)

        ### YOUR IMPLEMENTATION END  ###

        # Calculate derivative of error E with respect to parameter w
        δEδw = np.zeros_like(w)

        ### YOUR IMPLEMENTATION START  ###

        # Per sample version
        # for i in range(n):
        #      δEδw_i = np.outer(x[i,:], δEδy[i,:])
        #      δEδw += δEδw_i

        ## Vectorized version
        δEδw = x.T.dot(δEδy)

        ### YOUR IMPLEMENTATION END  ###

        return δEδx, {"w":δEδw}
